function HOMISBOLTZ;
%
% HOMISBOLTZ program, by Pietro Asinari, 10 April 2010
%
% The problem consists in integrating the homogeneous Boltzmann 
% equation for a generic collisional kernel in case of isotropic 
% symmetry, by a deterministic direct method. Difficulties arise 
% from the multi-dimensionality of the collisional operator and 
% from satisfying the conservation of mass and energy (momentum 
% is trivial for this test case) as accurately as possible, in 
% order to preserve the long time dynamics.
%
% The solution is based on the method proposed by Aristov [1], 
% but with two substantial improvements: 
%   (a) the original problem is reformulated in terms of particle 
%       kinetic energy (this allows one to ensure exact particle 
%       number and energy conservation during microscopic collisions) 
%   (b) a DVM-like correction (where DVM stands for Discrete Velocity 
%       Model) is adopted for improving the relaxation rates (this 
%       allows one to satisfy exactly the conservation laws at 
%       macroscopic level, which is particularly important for 
%       describing the late dynamics in the relaxation towards 
%       the equilibrium). 
% Both these corrections make possible to derive very accurate 
% reference solutions for this test case.
%
% [1] V.V. Aristov, Direct Methods for Solving the Boltzmann 
%     Equation and Study of Nonequilibrium Flows, Kluwer Academic 
%     Publishers, 2001.
%    
% Copyright 2010 Pietro Asinari (Politecnico di Torino, IT)
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

close all
clear all
clc

%==============================================================%
% Input data
%==============================================================%

% COLLISION KERNEL (See Eq. (3) for details)
a  = 1; % [m] particle radius
CS = 50; % [m/s] characteristic mean particle velocity
% Specify the collision kernel of the adopted model
% THETA = 0, Constant Kernel Model (CKM)
% THETA = 1, Hard Sphere Model (HSM)
THETA = 1; % [-]

% BGK APPROXIMATION (See Section 2.8)
% Specify if the BGK approximation must be considered 
% (where BGK stands for Bhatnagar-Gross-Krook)
% BGK = 0, full collisional integral of the BE
% BGK = 1, BGK-like approximation
% BGK = 2, full BGK approximation
BGK = 0; % [-]

% INITIAL CONDITION (See Eq. (110) for details)
EM  = 5e3; % [m^2/s^2] maximum particle kinetic energy
G0  = 600; % [m/s]
f0  = 5e-4; % [s^3/m^6]
G00 = 35; % [m^2/s^2]

% SIMULATION TIME (relaxation time scales as (n*a^2*CS)^(-1)
sim = 1e-3; % [s] simulation time

% DISCRETIZATION of the ENERGY (sub-)SPACE (See Section 2.5)
% This choice determines the accuracy of the solution. However, 
% it also determines the computational effort, in particular 
% for generating the data structure B(i,j) (see Section 4.1).
% Generating the structure B(i,j) can be quite demanding,
% in particular because of the DVM correction. Hence three 
% data structures are provided (as examples) for M = 50, 100.
% The code searches for the corresponding binary data files 
% or it generates them (once for all), if necessary.
M = 100; % [-] number of subdivisions for particle kinetic energy 

% DISCRETIZATION of TIME (See Section 6 and Eq. (112))
% The number of time steps will determine the magnitude of the time
% step, which must be selected according to the CFL stability 
% condition (where CFL stands for CourantFriedrichsLewy), since 
% we are using an explicit scheme (Forward Euler integration).
T = 100; % [-] number of time steps

% MOMENTS (See Section 2.7 and Eq. (97))
MOM = [2,3,4,5,6,7,8,9]; % [-] moments to monitor in post-processing

DeltaE = EM/M; % [m^2/s^2]
Deltat = sim/T; % [s]
CSP    = CS*(sqrt(DeltaE)/CS)^THETA; % [m/s]
Factor = 2^((THETA+3)/2)*(pi*a)^2*CSP; % [m^3/s], see Eq. (51)

%==============================================================
% Initialization
%==============================================================

% Let us create the structure B(i,j), which is used for 
% computing the GAIN term of the Boltzmann equation.
% See Section 4.1 "Data structure storing the dimensionless 
% frequencies" for details.

% The ideal value of matrix A is computed as well according to
% Eq. (52). It is useful to compute the ideal A for evaluating 
% the equilibrium relaxation frequency (nu_E)_i given by Eq. (100)

% In the call to the subroutine creating (or loading) the
% data structure B(i,j) (see Section 4.1 for details), the 
% second parameter is the DVM option. The DVM option 
% refers to the DVM correction (DVM=0 implies no correction, 
% while DVM=1 implies the DVM correction). The default is to 
% use the DVM correction, but it may be changed for checking
% the effects on the moment dynamics (DVM=0 corresponds to the 
% original Aristov's method [1]).
if (THETA==0)
    B = B_Constant_Kernel_Model(M,1);
    IdealA = A_Constant_Kernel_Model(M);
elseif (THETA==1)
    B = B_Hard_Sphere_Model(M,1);
    IdealA = A_Hard_Sphere_Model(M);
end

% Let us create the matrix A, which is used for describing 
% the LOSS term in the Boltzmann equation. By Eq. (86) or Eq. (96)
% (the latter in case of DVM correction), let us exclude the 
% post-collision velocities, which fall out of the adopted 
% descretization.

for i=1:M
    for j=1:M
        A(i,j) = sum(B(i,j).value(:));
    end
end

% Initial distribution function, see Eq. (110)
for i=1:M
    E(i)  = (i-1/2)*DeltaE;
    f(i)  = f0*exp(-(sqrt(E(i))-sqrt(G0))^2/G00);
end

% Storing the transient solution of the distribution function.
% This quantity can be easily made available out of the main 
% subroutine for future use, e.g. [fstored] = HOMISBOLTZ;
fstored = f;

% Recursive tuning of the equilibrium condition: see 
% Section 2.7 and Eqs. (98, 99) for details. This is an 
% essential feature for recovering the correct dynamics.
fe = Equilibrium(f,E,DeltaE);
[n,e,kT] = Thermodynamics(fe,DeltaE);
nue   = Nu_Equilibrium(Factor,DeltaE,E,fe,IdealA); % Eq. (100)
nuBGK = nue(1); % approximation: see after Eq. (104) 
% Moments at equilibrium depend on the quantity n (number
% density) and e (macroscopic kinetic energy), which are 
% constant in this problem. Hence the equilibrium moments are
% constant as well.
for q = 1:size(MOM,2)
    moments_eq(q) = Phi(fe,DeltaE,MOM(q));
end

%==============================================================
% Main Loop (See Section 4.2)
%==============================================================

for t = 1:(T+1)

    for i = 1:M
        N(i)  = 0;
        nu(i) = 0;
        
        for j = 1:M
            % GAIN term
            Psi(i,j) = 0;
            for m = 1:B(i,j).howmany
                k = B(i,j).k(m);
                l = B(i,j).l(m);
                Bijkm = B(i,j).value(m);
                Psi(i,j) = Psi(i,j)+f(k)*f(l)*Bijkm;
            end
            N(i)  = N(i)+Factor*DeltaE^(3/2)*(j-1/2)^(1/2)*Psi(i,j);
            
            % LOSS term
            nu(i) = nu(i)+Factor*DeltaE^(3/2)*(j-1/2)^(1/2)*f(j)*A(i,j);
        end
        
        if (BGK==0) % full collisional integral of the BE
            Q(i) = N(i)-nu(i)*f(i);
        elseif (BGK==1) % BGK-like approximation
            Q(i) = nue(i)*(fe(i)-f(i)); % See Eq. (101)
        elseif (BGK==2) % BGK approximation
            Q(i) = nuBGK*(fe(i)-f(i)); % See Eq. (104)
        end
    end

    % Computing (non-conserved) moments and their time rates
    time(t) = (t-1)*Deltat;
    for q = 1:size(MOM,2)        
        moments(t,q)     = Phi(f,DeltaE,MOM(q));
        dmoments_dt(t,q) = Phi(Q,DeltaE,MOM(q));
    end

    % Forward Euler integration rule (explicit). This simple 
    % integration rule leads to a CFL stability condition. See
    % Eq. (112) for details.
    f = f+Deltat.*Q;
    
    % Storing the transient solution of the distribution function.
    % This quantity can be easily made available out of the main 
    % subroutine for future use, e.g. [fstored] = HOMISBOLTZ;    
    fstored = [fstored;f];
    
    % Command line Post-processing
    checkR = Phi(f,DeltaE,0);
    checkE = Phi(f,DeltaE,1);
    checkM = Phi(f,DeltaE,2);
    pr = 'ITER: %2i/%i, DENSITY(p=0): %i, ENERGY(p=1): %i, MOM(p=2): %i';
    disp(sprintf(pr,t-1,T,checkR,checkE,checkM));

end

%==============================================================
% Plot Post-processing
%==============================================================

for q = 1:size(MOM,2)
    % Relaxation rate, see Eq. (113)
    rates(:,q)  = ( moments(:,q)-moments_eq(q) )/moments_eq(q);

    % Normalized relaxation rate
    ratesn(:,q) = rates(:,q)./rates(1,q);

    % Effective relaxation frequency, see Eq. (114)
    numom(:,q)  = dmoments_dt(:,q)./( moments_eq(q)-moments(:,q) );
end

figure(1) % corresponds to Fig. 1 in the paper
semilogx(E,fstored(1,:),'b-o'),grid on,hold on
semilogx(E,fe,'k-s'),grid on,hold on
semilogx(E,fstored(2:(T+1),:)','r'),grid on,hold on
axis tight
legend('Initial Condition','Equilibrium Condition')
xlabel('E, particle kinetic energy [m^2/s^2]')
ylabel('Distribution Function f(E) [s^3/m^6]')
title('Distribution Function Dynamics')

figure(2) % corresponds to Fig. 2 in the paper
plot(time,rates),grid on,hold on
axis tight
legend(num2str(MOM'))
xlabel('Time [s]')
ylabel('Moments Relaxation R_p [-]')
title('Macroscopic Moments Dynamics')

figure(3) % corresponds to Fig. 3 in the paper
plot(time,ratesn),grid on,hold on
axis tight
legend(num2str(MOM'))
xlabel('Time [s]')
ylabel('Normalized Moments Relaxation R_p(t)/R_p(0) [-]')
title('Normalized Macroscopic Moments Dynamics')

figure(4) % corresponds to Fig. 4 in the paper
plot(time,numom./nuBGK),grid on,hold on
axis tight
legend(num2str(MOM'))
xlabel('Time [s]')
ylabel('Normalized Relaxation Frequency \nu_p/\nu_{BGK} [-]')
title('Normalized Relaxation Frequency')

return

%**************************************************************%
%************************** Moments ***************************%
%**************************************************************%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Phip = Phi(f,DeltaE,p)
% This subroutine computes the moment of order p of the 
% discrete distribution function f. See Eq. (97).

M   = size(f,2);
Ep  = (([1:M]-1/2).*DeltaE).^(p+1/2);
fEp = f.*Ep;

Phip = 4*pi*sqrt(2)*DeltaE*sum(fEp);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [n,e,kT] = Thermodynamics(f,DeltaE)
% This subroutine computes some thermodynamic quantities of 
% the distribution function f. See the discussion at the end 
% of the Section 2.3.

n  = Phi(f,DeltaE,0);
ne = Phi(f,DeltaE,1);
e  = ne/n;
kT = 2*e/3; % monatomic gases

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function fe = Equilibrium(f,E,DeltaE)
% This subroutine computes the equilibrium distribution 
% function by means of a recursive fitting of the prescribed
% number density and kinetic energy. See the discussion at 
% the end of Section 2.7 and, in particular, Eqs. (98, 99).

[nI,eI,kTI] = Thermodynamics(f,DeltaE); % Target values
n   = nI;  % Guessed values
kT  = kTI; % Guessed values
M   = size(f,2);

Err = 2;
while Err>(10*eps)
    for i=1:M
        fe(i) = n/(2*pi*kT)^(3/2)*exp(-E(i)/kT);
    end
    [nr,er,kTr] = Thermodynamics(fe,DeltaE); % Real values
    n   = n+(nI-nr)/10;    % Update values
    kT  = kT+(kTI-kTr)/10; % Update values
    Err = abs(nr-nI)/nI+abs(kTr-kTI)/kTI;
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function nue = Nu_Equilibrium(F,DeltaE,E,fe,A)
% This subroutine computes the equilibrium relaxation 
% frequency (nu_E)_i according to Eq. (100). It uses the
% ideal matrix A, as described by Eq. (52).

M = size(fe,2);
for i=1:M
    nue(i) = 0;
    for j=1:M
        nue(i) = nue(i)+A(i,j)*sqrt(E(j))*fe(j);
    end
    nue(i) = F*DeltaE*nue(i);
end

return

%**************************************************************%
%**************** Constant_Kernel_Model ***********************%
%**************************************************************%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function A = A_Constant_Kernel_Model(M)
% This subroutine computes the ideal matrix A, as described 
% by Eq. (52).

for i=1:M
    for j=1:M
        A(i,j) = 4;
    end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function B = B_Constant_Kernel_Model(M,DVM)
% This subroutine searches for the binary data files storing the 
% data structures B(i,j) (see Section 4.1 for details). If this 
% files exists, it is loaded. Otherwise, the creator subroutine 
% of the data structure is called and the corresponding binary 
% data file is generated for future use (see Section 5 for details).

switch DVM
    case 0
        name = strcat('CKM_Structure_B_nodes_',num2str(M),'.mat');
        if exist(name)
            load(name);
        else
            B = B_Constant_Kernel_Model_Creator(M,DVM);
            name = strcat('CKM_Structure_B_nodes_',num2str(M),'.mat');
            save(name,'B');
        end
    case 1
        name = strcat('CKM_Structure_B_DVM_nodes_',num2str(M),'.mat');
        if exist(name)
            load(name);
        else
            B = B_Constant_Kernel_Model_Creator(M,DVM);
            name = strcat('CKM_Structure_B_DVM_nodes_',num2str(M),'.mat');
            save(name,'B');
        end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function B = B_Constant_Kernel_Model_Creator(M,DVM)
% This subroutine computes the data structure B(i,j), which is used
% for describing the GAIN term in the Boltzmann equation (see
% Section 4.1 for details). The parameter DVM refers to the DVM
% correction (DVM=0 implies no correction, while DVM=1 implies 
% the DVM correction). 

nEmin = 1/2;
nEmax = M-1/2;

for i=1:M
    pr = 'CKM, computing structure B: progress %3.2f %%';
    disp(sprintf(pr,i/M*100));
    nEi = i-1/2; % normalized Ei /Delta E
    
    for j=1:M
        nEj = j-1/2; % normalized Ej /Delta E    
        lambda = 0;

        for k=1:M;
            nEk = k-1/2; % normalized Ej /Delta E
            nEl = nEi+nEj-nEk; % energy conservation
            
            if ( (nEl>=nEmin) && (nEl<=nEmax) )
                lambda = lambda+1;
                
                BB(lambda,1) = k;
                l = round(nEl+1/2);
                BB(lambda,2) = l;
                
                nEkm = k-1;
                nEkp = k;
                BB(lambda,3) = nEkm;
                BB(lambda,4) = nEkp;
                
                Cp = C_Constant_Kernel_Model(nEi,nEj,nEkp);
                Cm = C_Constant_Kernel_Model(nEi,nEj,nEkm);
                BB(lambda,5) = Cp-Cm;
            end
        end
        
        B(i,j).howmany = size(BB,1);
        B(i,j).k       = BB(:,1);
        B(i,j).l       = BB(:,2);
        B(i,j).nEkm    = BB(:,3);
        B(i,j).nEkp    = BB(:,4);
        B(i,j).value   = BB(:,5);
        
        clear BB
    end
end

% DVM Correction, See Eqs. (93, 95) for details
if DVM
    for i=1:M
        pr = 'CKM, computing DVM correction: progress %3.2f %%';
        disp(sprintf(pr,i/M*100));
        for j=1:M
            for lam=1:B(i,j).howmany
                k = B(i,j).k(lam);
                l = B(i,j).l(lam);

                Permijkl  = EnergyPerms([i j k l]);
                Gammaijkl = 0;
                 for lamp=1:size(Permijkl,1)
                    ip = Permijkl(lamp,1);
                    jp = Permijkl(lamp,2);
                    kp = Permijkl(lamp,3);
                    lp = Permijkl(lamp,4);
                    Bijkl = Read_B(B,ip,jp,kp,lp);
                    Gammaijkl = Gammaijkl+sqrt((ip-1/2)*(jp-1/2))*Bijkl;
                end
                Gammaijkl = Gammaijkl/size(Permijkl,1);
                for lamp=1:size(Permijkl,1)
                    ip = Permijkl(lamp,1);
                    jp = Permijkl(lamp,2);
                    kp = Permijkl(lamp,3);
                    lp = Permijkl(lamp,4);
                    Bijkl = Gammaijkl/sqrt((ip-1/2)*(jp-1/2));
                    B = Write_B(B,ip,jp,kp,lp,Bijkl);
                end
            end
        end
    end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function C = C_Constant_Kernel_Model(nEi,nEj,nEkpm)
% This subroutine computes the function C(Ek) according to 
% Eq. (67). In particular, Case 1 corresponds to C_1(0) given by
% Eq. (71), Case 2 to C_2(0) given by Eq. (76) and Case 3 to
% C_3(0) given by Eq. (80).

if ( nEi<=nEj )
    if ( nEkpm<=nEi ) % Case 1
        C = 2*sqrt(nEkpm)/sqrt(nEj)+(nEi-nEkpm)/sqrt(nEi*nEj)*...
            log((sqrt(nEi)-sqrt(nEkpm))/(sqrt(nEi)+sqrt(nEkpm)));
    elseif ( ( nEi<=nEkpm ) && ( nEkpm<=nEj ) ) % Case 2
        C = 2*sqrt(nEkpm)/sqrt(nEj)+(nEkpm-nEi)/sqrt(nEi*nEj)*...
            log((sqrt(nEkpm)+sqrt(nEi))/(sqrt(nEkpm)-sqrt(nEi)));
    elseif ( nEj<=nEkpm ) % Case 3
        C = 4-2*sqrt((nEi+nEj-nEkpm)/nEi)...
            -2*(nEi-nEkpm)/sqrt(nEi*nEj)*...
            log((sqrt(nEj)+sqrt(nEi+nEj-nEkpm))/...
            sqrt(nEj-(nEi+nEj-nEkpm)));
    end
else
    if ( nEkpm<=nEj ) % Case 1
        C = 2*sqrt(nEkpm)/sqrt(nEi)+(nEj-nEkpm)/sqrt(nEj*nEi)*...
            log((sqrt(nEj)-sqrt(nEkpm))/(sqrt(nEj)+sqrt(nEkpm)));
    elseif ( ( nEj<=nEkpm ) && ( nEkpm<=nEi ) ) % Case 2
        C = 2*sqrt(nEkpm)/sqrt(nEi)+(nEkpm-nEj)/sqrt(nEj*nEi)*...
            log((sqrt(nEkpm)+sqrt(nEj))/(sqrt(nEkpm)-sqrt(nEj)));
    elseif ( nEi<=nEkpm ) % Case 3
        C = 4-2*sqrt((nEj+nEi-nEkpm)/nEj)...
            -2*(nEj-nEkpm)/sqrt(nEj*nEi)*...
            log((sqrt(nEi)+sqrt(nEj+nEi-nEkpm))/...
            sqrt(nEi-(nEj+nEi-nEkpm)));
    end
end

return

%**************************************************************%
%****************** Hard_Sphere_Model *************************%
%**************************************************************%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function A = A_Hard_Sphere_Model(M)
% This subroutine computes the ideal matrix A, as described 
% by Eq. (52).

for i=1:M
    nEi  = i-1/2; % normalized Ei /Delta E

    for j=1:M
        nEj  = j-1/2; % normalized Ej /Delta E
        
        if ( nEi<=nEj )
            A(i,j) = 2*sqrt(nEj)+2/3*nEi/sqrt(nEj);
        else
            A(i,j) = 2*sqrt(nEi)+2/3*nEj/sqrt(nEi);
        end
    end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function B = B_Hard_Sphere_Model(M,DVM)
% This subroutine searches for the binary data files storing the 
% data structures B(i,j) (see Section 4.1 for details). If this 
% files exists, it is loaded. Otherwise, the creator subroutine 
% of the data structure is called and the corresponding binary 
% data file is generated for future use (see Section 5 for details).

switch DVM
    case 0
        name = strcat('HSM_Structure_B_nodes_',num2str(M),'.mat');
        if exist(name)
            load(name);
        else
            B = B_Hard_Sphere_Model_Creator(M,DVM);
            name = strcat('HSM_Structure_B_nodes_',num2str(M),'.mat');
            save(name,'B');
        end
    case 1
        name = strcat('HSM_Structure_B_DVM_nodes_',num2str(M),'.mat');
        if exist(name)
            load(name);
        else
            B = B_Hard_Sphere_Model_Creator(M,DVM);
            name = strcat('HSM_Structure_B_DVM_nodes_',num2str(M),'.mat');
            save(name,'B');
        end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function B = B_Hard_Sphere_Model_Creator(M,DVM)
% This subroutine computes the data structure B(i,j), which is used
% for describing the GAIN term in the Boltzmann equation (see
% Section 4.1 for details). The parameter DVM refers to the DVM
% correction (DVM=0 implies no correction, while DVM=1 implies 
% the DVM correction). 

nEmin = 1/2;
nEmax = M-1/2;

for i=1:M
    pr = 'HSM, computing structure B: progress %3.2f %%';
    disp(sprintf(pr,i/M*100));
    nEi = i-1/2; % normalized Ei /Delta E
    
    for j=1:M
        nEj = j-1/2; % normalized Ej /Delta E    
        lambda = 0;

        for k=1:M;
            nEk = k-1/2; % normalized Ej /Delta E
            nEl = nEi+nEj-nEk; % energy conservation
            
            if ( (nEl>=nEmin) && (nEl<=nEmax) )
                lambda = lambda+1;
                
                BB(lambda,1) = k;
                l = round(nEl+1/2);
                BB(lambda,2) = l;
                
                nEkm = k-1;
                nEkp = k;
                BB(lambda,3) = nEkm;
                BB(lambda,4) = nEkp;
                
                Cp = C_Hard_Sphere_Model(nEi,nEj,nEkp);
                Cm = C_Hard_Sphere_Model(nEi,nEj,nEkm);
                BB(lambda,5) = Cp-Cm;
            end
        end
        
        B(i,j).howmany = size(BB,1);
        B(i,j).k       = BB(:,1);
        B(i,j).l       = BB(:,2);
        B(i,j).nEkm    = BB(:,3);
        B(i,j).nEkp    = BB(:,4);
        B(i,j).value   = BB(:,5);
        
        clear BB
    end
end

% DVM Correction, See Eqs. (93, 95) for details
if DVM
    for i=1:M
        pr = 'HSM, computing DVM correction: progress %3.2f %%';
        disp(sprintf(pr,i/M*100));
        for j=1:M
            for lam=1:B(i,j).howmany
                k = B(i,j).k(lam);
                l = B(i,j).l(lam);

                Permijkl  = EnergyPerms([i j k l]);
                Gammaijkl = 0;
                 for lamp=1:size(Permijkl,1)
                    ip = Permijkl(lamp,1);
                    jp = Permijkl(lamp,2);
                    kp = Permijkl(lamp,3);
                    lp = Permijkl(lamp,4);
                    Bijkl = Read_B(B,ip,jp,kp,lp);
                    Gammaijkl = Gammaijkl+sqrt((ip-1/2)*(jp-1/2))*Bijkl;
                end
                Gammaijkl = Gammaijkl/size(Permijkl,1);
                for lamp=1:size(Permijkl,1)
                    ip = Permijkl(lamp,1);
                    jp = Permijkl(lamp,2);
                    kp = Permijkl(lamp,3);
                    lp = Permijkl(lamp,4);
                    Bijkl = Gammaijkl/sqrt((ip-1/2)*(jp-1/2));
                    B = Write_B(B,ip,jp,kp,lp,Bijkl);
                end
            end
        end
    end
end

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function C = C_Hard_Sphere_Model(nEi,nEj,nEkpm)
% This subroutine computes the function C(Ek) according to 
% Eq. (67). In particular, Case 1 corresponds to C_1(1) given by
% Eq. (72), Case 2 to C_2(1) given by Eq. (77) and Case 3 to
% C_3(1) given by Eq. (81).

if ( nEi<=nEj )
    if ( nEkpm<=nEi ) % Case 1
        C = 4/3*nEkpm^(3/2)/( sqrt(nEi)*sqrt(nEj) );
    elseif ( ( nEi<=nEkpm ) && ( nEkpm<=nEj ) ) % Case 2
        C = 2*(3*nEkpm-nEi)/(3*sqrt(nEj));
    elseif ( nEj<=nEkpm ) % Case 3
        C = 2*sqrt(nEj)+2/3*nEi/sqrt(nEj)...
            -4/3*( nEi+nEj-nEkpm )^(3/2)/( sqrt(nEi)*sqrt(nEj) );
    end
else
    if ( nEkpm<=nEj ) % Case 1
        C = 4/3*nEkpm^(3/2)/( sqrt(nEj)*sqrt(nEi) );
    elseif ( ( nEj<=nEkpm ) && ( nEkpm<=nEi ) ) % Case 2
        C = 2*(3*nEkpm-nEj)/(3*sqrt(nEi));
    elseif ( nEi<=nEkpm ) % Case 3
        C = 2*sqrt(nEi)+2/3*nEj/sqrt(nEi)...
            -4/3*( nEj+nEi-nEkpm )^(3/2)/( sqrt(nEj)*sqrt(nEi) );
    end
end

return

%**************************************************************%
%******************* Models Utilities *************************%
%**************************************************************%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Bijkl = Read_B(B,i,j,k,l)
% This subroutine reads the value B_{ij}^{kl} in the data 
% structure B(i,j) (see Section 4.1 for details).

lam = find(B(i,j).k(:)==k);
Bijkl = B(i,j).value(lam);

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function B = Write_B(B,i,j,k,l,Bijkl)
% This subroutine writes the value Bijkl in the cell B_{ij}^{kl} 
% of the data structure B(i,j) (see Section 4.1 for details).

lam = find(B(i,j).k(:)==k);
B(i,j).value(lam) = Bijkl;

return

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function M = EnergyPerms(v)
% This subroutine computes the permutations of the indexes 
% v(1), v(2), v(3), v(4) which ensure the energy conservation,
% namely v(1)+v(2)=v(3)+v(4). There are 8 combinations, if all 
% indexes are different, which reduce to 4 if v(1)=v(2) or 
% v(3)=v(4) and obviously to 1 if v(1)=v(2)=v(3)=v(4).
% See the discussion in Section 2.6, after Eq. (92) for details.

if (v(1)==v(2)==v(3)==v(4))
    M = v;
elseif (v(1)==v(2))
    % 4 cases
    M = [
     v(3),     v(4),     v(1),     v(2);
     v(1),     v(2),     v(3),     v(4);
     v(1),     v(2),     v(4),     v(3);
     v(4),     v(3),     v(1),     v(2)];
elseif (v(3)==v(4))
    % 4 cases
    M = [
     v(2),     v(1),     v(3),     v(4);
     v(3),     v(4),     v(2),     v(1);
     v(3),     v(4),     v(1),     v(2);
     v(1),     v(2),     v(3),     v(4)];
else
    % 8 cases
    M = [
     v(4),     v(3),     v(2),     v(1);
     v(4),     v(3),     v(1),     v(2);
     v(3),     v(4),     v(2),     v(1);
     v(3),     v(4),     v(1),     v(2);
     v(2),     v(1),     v(4),     v(3);
     v(2),     v(1),     v(3),     v(4);
     v(1),     v(2),     v(3),     v(4);
     v(1),     v(2),     v(4),     v(3)];
end

return
